/********************************************************************
 * FILE: beadstring.c
 * AUTHOR: William Stafford Noble Charles E. Grant
 * CREATE DATE: 06-23-2007
 * PROJECT: MHMM
 * COPYRIGHT: 2007, WNG
 * DESCRIPTION: Given MEME output, create a motif-based linear HMM and
 *              use it to search a sequence database for motif clusters.
 ********************************************************************/
#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "matrix.h"         // Matrix functions
#include "alphabet.h"       // Amino acid and nucleotide alphabets
#include "build-hmm.h"      // Construct an HMM in memory. 
#include "dp.h"             // Dynamic programming algorithms
#include "fasta-io.h"       // Reading and writing FASTA files
#include "fitevd.h"         // Functions for fitting an extreme value dist
#include "log-hmm.h"        // Convert an HMM to or from log form. 
#include "match.h"          // Find sequence to model match from traceback
#include "metameme.h"       // Global Meta-MEME functions. 
#include "meme-io.h"        // Read motifs from MEME output. 
#include "mhmm.h"          // Scan a sequence db with HMM.
#include "mhmm-state.h"     // Data structure for HMMs. 
#include "mhmms.h"          // Scan a sequence db with HMM.
#include "motif.h"          // Data structure for holding motifs. 
#include "order.h"          // Linear motif order and spacing. 
#include "pssm.h"           // Tools for position specific scoring matrix
#include "read-mhmm.h"      // Functions for reading HMM from file. 
#include "simple-getopt.h"  // Command line processing
#include "string-list.h"    // List of strings (for motif IDs).
#include "utils.h"          // Generic utilities. 
#include "write-mhmm.h"     // Functions for producing MHMM output. 

#define SLOP 1E-5

// Marker for end transition.
extern char* END_TRANSITION;

VERBOSE_T verbosity = NORMAL_VERBOSE;

enum path_values { SINGLE, ALL };

// Structure for tracking beadstring command line parameters.
typedef struct options {

  char* meme_filename;    /* Name of file containg motifs. */
  char* seq_filename;     /* Name of file containg sequences. */
  char* bg_filename;      /* Name of file file containg background freq. */
  char* score_filename;
  char* model_filename;   /* Name of output file for HMM. */
  char* order_string;     /* Motif order and spacing. */

  BOOLEAN_T allow_weak_motifs;
  BOOLEAN_T fim;              /* Represent spacers as free insertion modules? */
  BOOLEAN_T local_scoring;
  BOOLEAN_T motif_scoring;
  BOOLEAN_T perform_search;
  BOOLEAN_T synth;
  BOOLEAN_T zselo;

  int       max_seq;
  int       pam_distance;
  int       paths;
  int       spacer_states;

  PROB_T  motif_e_threshold;   /* e-value threshold for motif inclusion. */
  PROB_T  motif_p_threshold;   /* p-value threshold for motif occurences. */
  PROB_T  e_threshold;   /* e-value threshold for reporting cluster. */
  PROB_T  p_threshold;   /* Turns on p-value scoring and sets zer point. */
  double  gap_extend;    /* penalty for extending a gap */
  double  gap_open;      /* penalty for opening a gap */
  double  motif_pseudo;  /* pseudocount  for motif frequencies */
  double  spacer_pseudo; /* Spacer (self-loop) pseudocount. */
  double  trans_pseudo;  /* Transition pseudocount. */
  double  progress;      /* interval between progress reports */

  STRING_LIST_T* requested_motifs; /* Indices of requested motifs. */
} OPTIONS_T;


/************************************************************************
 * Fix list of transitions in
 ************************************************************************/
void fix_trans_in(MHMM_T* the_hmm) {

  int i_row, i_col;
  int n = 0;
  MATRIX_T *trans = the_hmm->trans;
  n = trans->num_rows;

  //
  // Visit the transition matrix cells just once each
  // to update trans_in arrays.
  // This is quadratic in n. 
  //
  for (i_col = 0; i_col < n; i_col++) {
    for (i_row = 0; i_row < n; i_row++) {
      double p = get_matrix_cell(i_row, i_col, trans);	// The transition probability.
      int m = 0;
      if (!is_zero(p, FALSE)) {
        MHMM_STATE_T * in_state = &(the_hmm->states[i_col]);
        assert(m < in_state->ntrans_in);
        set_array_item(m, p, in_state->trans_in);
        m++;
      }
    } // col
  } // row

} // compute_ins_and_outs

/*************************************************************************
 *  Build a linear HMM model for the motifs
 *************************************************************************/
MHMM_T* build_linear_model(struct options* options) {

  /* Data structures. */
  int num_motifs = 0;         /* The number of motifs in the model. */
  MOTIF_T  motifs[2 * MAX_MOTIFS]; /* The motifs. */
  BOOLEAN_T has_reverse_strand = FALSE; /* MEME file contained both strands */
  ARRAY_T* background = NULL;     /* Background probs for alphabet. */
  ORDER_T* order_spacing = NULL;  /* Linear HMM order and spacing. */
  MATRIX_T* transp_freq = NULL;    /* Matrix of inter-motif transitions freqs. */
  MATRIX_T* spacer_ave = NULL;     /* Matrix of average spacer lengths. */
  MHMM_T* the_hmm = NULL;        /* The HMM being constructed. */
  STRING_LIST_T* motif_occurrences = NULL; /* Strings describing occurrences 
                                             of motifs */

  /* Gotta have positive number of spacer states. */
  if (options->spacer_states <= 0) {
    die("Negative spacer length (%d).\n", options->spacer_states);
  }

  /* Make sure motifs weren't selected redundantly. */
  if ((get_num_strings(options->requested_motifs) != 0) 
      && (options->order_string != NULL)) {
    die("Can't combine --motif and --order options.");
  }

  // Parse the order string. 
  order_spacing = create_order(options->order_string);

  /**********************************************
   * READING THE MOTIFS
   **********************************************/

  BOOLEAN_T read_file = FALSE;

  /* Asssume MEME XML file. */
  read_file = read_meme_xml_file(
     options->meme_filename,
     options->bg_filename,
     options->motif_pseudo,
     &num_motifs,
     motifs,
     &motif_occurrences,
     &has_reverse_strand,
     &background
  );

  if (read_file == FALSE) {
    // Wasn't MEME XML, try old file format.
    read_meme_file(
       options->meme_filename,
       options->bg_filename,
       options->motif_pseudo,
       &num_motifs,
       motifs,
       &motif_occurrences,
       &has_reverse_strand,
       &background
    );
  }

  process_raw_motifs_for_model(
       &num_motifs,
       motifs,
       motif_occurrences,
       options->requested_motifs,
       has_reverse_strand,
       FALSE, // keep unused motifs
       options->motif_p_threshold,
       options->motif_e_threshold,
       0.0, // complexity threshold
       &order_spacing,
       &transp_freq,
       &spacer_ave,
       options->trans_pseudo,
       options->spacer_pseudo
  );

  /**********************************************
   * BUILDING THE HMM
   **********************************************/

  /* Build the motif-based HMM. */

  if (order_spacing != NULL) {
    reorder_motifs(order_spacing, &num_motifs, motifs);
  }
  else {
    die("No order specified for the motifs.\n"
        "For the linear model the motif file must contain motif occurence\n" 
        "data or the motif order must be specified using the --order option.");
  }

  build_linear_hmm(
    background,
    order_spacing,
    options->spacer_states,
    motifs,
    num_motifs, 
    options->fim,
    &the_hmm
  );
  fix_trans_in(the_hmm);

  free_string_list(motif_occurrences);
  free_array(background);
  free_order(order_spacing);
  free_matrix(transp_freq);
  free_matrix(spacer_ave);
  int i_motif = 0;
  for (i_motif = 0; i_motif < num_motifs; i_motif++) {
    free_motif(&(motifs[i_motif]));
  }

  // Add some global information.
  copy_string(&(the_hmm->motif_file), options->meme_filename);

  return the_hmm;

}


void search_with_model(
 MHMM_T* the_mhmm, 
 BOOLEAN_T* got_evd,
 OPTIONS_T *options
) {

  BOOLEAN_T viterbi = FALSE;
  BOOLEAN_T use_pvalues = FALSE;
  BOOLEAN_T both_strands = FALSE;
  EVD_SET evd_set;      // EVD data structure 

  /***********************************************
   * Set up the model.
   ***********************************************/

  // Figure out what kind of scoring to do.
  if (options->paths == SINGLE) {
    viterbi = TRUE;
  } else {
    viterbi = FALSE;
  }

  // Force p-value scoring if p-value threshold given.
  if (options->p_threshold != 0) {
    use_pvalues = TRUE;
  }

  // Check p-threshold is in range [0<p<=1].
  if (use_pvalues && (options->p_threshold <= 0 || options->p_threshold > 1)) {
    die("You must specify p-thresh in the range [0<p<=1]\n");
  }

  // Set the PAM distance.
  if (options->pam_distance == -1) {
    options->pam_distance = (which_alphabet() == PROTEIN_ALPH) 
      ? DEFAULT_PROTEIN_PAM : DEFAULT_DNA_PAM;
  }
  int beta = (which_alphabet() == PROTEIN_ALPH) ? 10 : 1;

  // Convert the model to log space.
  MHMM_T*   the_log_hmm;      // The HMM, with probs converted to logs. 
  convert_to_from_log_hmm(
    TRUE, // Convert to logs.
    options->zselo,
    options->gap_open,
    options->gap_extend,
    the_mhmm->background,
    options->score_filename,
    options->pam_distance,
    beta,
    the_mhmm, 
    &the_log_hmm
  );

  // Set up PSSM matrices if doing motif_scoring
  // and pre-compute motif p-values if using p-values.
  // Set up the hot_states list.
  set_up_pssms_and_pvalues(
    options->motif_scoring,
    options->p_threshold,
    both_strands,
    options->allow_weak_motifs,
    the_log_hmm
  );

  //
  // Set up for computing score distribution.
  //
  SCORE_SET *score_set = set_up_score_set(
    options->p_threshold,   
    -1,    // dp_threshold, 
    -1,    // max_gap
    FALSE, // negatives_only
    the_log_hmm
  );

  /***********************************************
   * Search the database one sequence at a time.
  ***********************************************/

  // Open the file for reading. 
  FILE* seq_file = NULL;
  if (open_file(
        options->seq_filename, 
        "r", 
        TRUE, 
        "sequence", 
        "sequences", 
        &seq_file
      ) == 0) {
    exit(1);
  }

  double start_time = myclock();
  double old_loop_time = start_time;
  double seconds_since_last_report = 0.0;

  SEQ_T* sequence = NULL;        // Sequence to search against. 
  MATRIX_T* motif_score_matrix = NULL;  // Number of motifs x sequence length.
  MATRIX_T* trace_matrix = NULL; // Traceback for Viterbi.
  MATCH_T*  this_match = NULL; // This sequence-to-model match.
  MATRIX_T* dp_matrix = NULL; // Dynamic programming matrix. 
  int dp_rows = 0;            // Size of the DP matrix. 
  int dp_cols = 0;
  int num_seqs = 0;

  this_match = allocate_match(); // This sequence-to-model match.
  while (read_one_fasta(seq_file, MAX_SEQ, &sequence)) {

    num_seqs++;

    // Keep track of total database size for E-value calculation.
    score_set->total_length += get_seq_length(sequence);

    // Let the user know what's going on.
    if (verbosity > NORMAL_VERBOSE) {
      fprintf(
        stderr, 
        "Scoring %s (length=%d).\n", 
        get_seq_name(sequence),
        get_seq_length(sequence)
      );
    }

    // Convert the sequence to alphabet-specific indices. 
    prepare_sequence(sequence);
    assert(get_seq_char(get_seq_length(sequence), sequence) == '\0');

    /* Allocate the dynamic programming matrix. Rows correspond to
       states in the model, columns to positions in the sequence. */
    if ((dp_rows < the_log_hmm->num_states) 
      || (dp_cols < get_seq_length(sequence))) {
      free_matrix(dp_matrix);
      free_matrix(trace_matrix);
      if (dp_rows < the_log_hmm->num_states) {
        dp_rows = the_log_hmm->num_states;
      }
      if (dp_cols < get_seq_length(sequence)) {
        dp_cols = get_seq_length(sequence);
      }
      // (Add one column for repeated match algorithm.)
      dp_matrix = allocate_matrix(dp_rows, dp_cols + 1);
      trace_matrix = allocate_matrix(dp_rows, dp_cols + 1);
    }

    // Compute the motif scoring matrix.
    if (options->motif_scoring) {
      motif_score_matrix = allocate_matrix(
        the_log_hmm->num_motifs,
        get_seq_length(sequence)
      );
      compute_motif_score_matrix(
        use_pvalues,
        options->p_threshold,
        get_int_sequence(sequence),
        get_seq_length(sequence),
        the_log_hmm,
        &motif_score_matrix
      );
    }

    // Perform the appropriate type of dynamic programming.
    if (viterbi) {
      viterbi_algorithm(
        options->local_scoring,
        get_int_sequence(sequence),
        get_seq_length(sequence),
        the_log_hmm,
        TRUE,  // Save Viterbi path?
        motif_score_matrix,
        dp_matrix,
        trace_matrix,
        this_match
      );
    } else {
      forward_algorithm(
        options->local_scoring,
        get_int_sequence(sequence),
        get_seq_length(sequence),
        the_log_hmm,
        //motif_score_matrix, // FIXME
        dp_matrix,
        this_match
      );
    }      

    // Store the score, ID, length and comment for later printing. 
    //assert(get_match_seq_length(this_match) == get_seq_length(sequence));
    //assert(get_match_seq_length(this_match) == get_seq_length(sequence));
    store_sequence(
      options->motif_scoring,
      FALSE, // Don't print in mhmmscan format.
      DEFAULT_OUTPUT_WIDTH, 
      sequence, 
      this_match, 
      options->e_threshold,
      0,    // no dp_threshold
      options->p_threshold,
      score_set,
      *got_evd, 
      evd_set,
      FALSE, // print fancy
      the_log_hmm,
      motif_score_matrix,
      FALSE // Store GFF?
    );

    /* Calculate the initial E-value distribution if the required
     * number of sequences has been saved.  This will allow the
     * descriptions of low-scoring sequences to not be stored.  The
     * distribution will be recomputed using all saved scores when all
     * sequences have been read.  */
    if (score_set->n == EVD_NUM_SEQS && *got_evd == FALSE) {
      evd_set = calc_distr(
        *score_set,   // Set of scores.
        use_pvalues, // Use exponential distribution?
        FALSE         // Use sequence E-values.
      );
      if (evd_set.n > 0) *got_evd = TRUE;
    }

    // Free the memory used by this sequence. 
    free_matrix(motif_score_matrix);
    free_seq(sequence);

    // If requested, periodically report progress
    if (options->progress > 0.0) {
      double new_loop_time = myclock();
      seconds_since_last_report = (new_loop_time - old_loop_time) / 1e6;
      if (seconds_since_last_report > options->progress) {
        double elapsed_seconds = (new_loop_time - start_time) / 1e6;
        old_loop_time = new_loop_time;
        fprintf(
          stderr,
          "%d sequences processed in %d seconds.\n", 
          num_seqs,
          (int) elapsed_seconds
        );
      }
    }
  }

  //double end_time = myclock();

  /***********************************************
   * Calculate the E-values and store them as the keys.
   ***********************************************/
  // Recalculate the EVD using all scores. 
  // If successful, calculate the E-values and
  // store them as keys.
  evd_set = calc_distr(
    *score_set,   // Set of scores.
    use_pvalues, // Use exponential distribution?
    FALSE         // Use sequence E-values.
  );
  if (evd_set.n > 0) {
    int q, t, N;
    q = 1;        // Ignore query "length".
    t = 0;        // Use stored target lengths.
    N = evd_set.non_outliers;    // Use number of non-outliers.
    evd_set.N = N;
    calc_evalues(&evd_set, N, q, t);
    *got_evd = TRUE;
  }
  evd_set.negatives_only = FALSE;  // Used real sequences.

  /* Tie up loose ends. */
  // myfree(evd_set.evds);
  free_mhmm(the_log_hmm);
  free_matrix(dp_matrix);
  free_matrix(trace_matrix);
  free_match(this_match);
  free(score_set->scores);
  myfree(score_set);
  fclose(seq_file);
}
  
/***********************************************************************
  Write search results to file
 ***********************************************************************/
void write_search_results(
  FILE* output_file, 
  MHMM_T* the_hmm, 
  BOOLEAN_T got_evd, 
  OPTIONS_T* options
) {

  /***********************************************
   * Print header information.
   ***********************************************/
  write_header(
    MHMMS_HEADER,
    the_hmm->description,
    the_hmm->motif_file,
    NULL,
    options->seq_filename,
    output_file
  );

  /***********************************************
   * Sort and print the results.
   ***********************************************/
  if (verbosity >= NORMAL_VERBOSE) {
    fprintf(stderr, "\nSorting the E-values.\n");
  }

  sort_and_print_scores(
    FALSE, // print fancy 
    TRUE,  // print_header, 
    got_evd,
    options->motif_scoring,
    FALSE, // Don't print in mhmmscan format.
    options->max_seq, // Maximum number of sequences to print
    DEFAULT_OUTPUT_WIDTH,
    options->e_threshold,
    TRUE, // Sort output 
    NULL, // FIXME: GFF output?
    output_file
  );

}

/***********************************************************************
  Process command line options
 ***********************************************************************/
static void process_command_line(
  int argc, 
  char* argv[], 
  OPTIONS_T *options
) {

  // Define command line options.
  cmdoption const option_list[] = {
    {"allow-weak-motifs", NO_VALUE},
    {"bgfile", REQUIRED_VALUE},
    {"e-thresh", REQUIRED_VALUE},
    {"fim", NO_VALUE},
    {"gap-extend", REQUIRED_VALUE},
    {"gap-open", REQUIRED_VALUE},
    {"global", NO_VALUE},
    {"maxseqs", REQUIRED_VALUE},
    {"model", REQUIRED_VALUE},
    {"motif", REQUIRED_VALUE},
    {"motif-e-thresh", REQUIRED_VALUE},
    {"motif-p-thresh", REQUIRED_VALUE},
    {"motif-pseudo", REQUIRED_VALUE},
    {"motif-scoring", NO_VALUE},
    {"no-search", NO_VALUE},
    {"nspacer", REQUIRED_VALUE},
    {"order", REQUIRED_VALUE},
    {"pam", REQUIRED_VALUE},
    {"paths", REQUIRED_VALUE},
    {"p-thresh", REQUIRED_VALUE},
    {"progress", REQUIRED_VALUE},
    {"score-file", REQUIRED_VALUE},
    {"spacer-pseudo", REQUIRED_VALUE},
    {"synth", NO_VALUE},
    {"trans-pseudo", REQUIRED_VALUE},
    {"verbosity", REQUIRED_VALUE},
    {"zselo", NO_VALUE},
  };
  int option_count = 27;
  int option_index = 0;

  // Define the usage message.
  char      usage[1000] = "";
  strcat(usage, "USAGE: beadstring [options] <motif file> <sequence file>\n");
  strcat(usage, "\n");
  strcat(usage, "   Options:\n");
  strcat(usage, "     --allow-weak-motifs\n");
  strcat(usage, "     --bgfile <background file>\n");
  strcat(usage, "     --e-thresh <float>\n");
  strcat(usage, "     --fim\n");
  strcat(usage, "     --gap-extend <float>\n");
  strcat(usage, "     --gap-open <float>\n");
  strcat(usage, "     --global\n");
  strcat(usage, "     --maxseqs <int>\n");
  strcat(usage, "     --model <model file>\n");
  strcat(usage, "     --motif <string>\n");
  strcat(usage, "     --motif-e-thresh <float>\n");
  strcat(usage, "     --motif-p-thresh <float>\n");
  strcat(usage, "     --motif-pseudo <float>\n");
  strcat(usage, "     --motif-scoring\n");
  strcat(usage, "     --motif-scoring\n");
  strcat(usage, "     --no-search\n");
  strcat(usage, "     --nspacer <int>\n");
  strcat(usage, "     --order <string>\n");
  strcat(usage, "     --pam <int>\n");
  strcat(usage, "     --paths single|all\n");
  strcat(usage, "     --p-thresh <float>\n");
  strcat(usage, "     --progress <float>\n");
  strcat(usage, "     --score-file <score file>\n");
  strcat(usage, "     --spacer-pseudo <float>\n");
  strcat(usage, "     --synth\n");
  strcat(usage, "     --trans-pseudo <float>\n");
  strcat(usage, "     --verbosity 1|2|3|4|5 (default=2)\n");
  strcat(usage, "     --zselo\n");
  strcat(usage, "\n");

  /* Make sure various options are set to NULL or defaults. */
  options->meme_filename = NULL;
  options->bg_filename = NULL;
  options->score_filename = NULL;
  options->seq_filename = NULL;
  options->model_filename = NULL;
  options->order_string = NULL;
  options->allow_weak_motifs = FALSE;
  options->fim = FALSE;
  options->local_scoring = TRUE;
  options->motif_scoring = FALSE;
  options->synth = FALSE;
  options->zselo = FALSE;
  options->max_seq = -1;
  options->pam_distance = -1;
  options->paths = SINGLE;
  options->spacer_states = DEFAULT_SPACER_STATES,
  options->gap_extend = -1.0;
  options->gap_open = -1.0;
  options->e_threshold = 0.0;
  options->p_threshold = 0.0;
  options->perform_search = TRUE;
  options->progress = 0.0;
  options->requested_motifs = new_string_list();
  options->motif_e_threshold = 0.0;
  options->motif_p_threshold = 0.0;
  options->motif_pseudo = DEFAULT_MOTIF_PSEUDO;
  options->spacer_pseudo = DEFAULT_SPACER_PSEUDO;
  options->trans_pseudo = DEFAULT_TRANS_PSEUDO;
  verbosity = 2;

  simple_setopt(argc, argv, option_count, option_list);

  // Parse the command line.
  while (1) { 
    int c = 0;
    char* option_name = NULL;
    char* option_value = NULL;
    const char * message = NULL;


    // Read the next option, and break if we're done.
    c = simple_getopt(&option_name, &option_value, &option_index);
    if (c == 0) {
      break;
    } else if (c < 0) {
      simple_getopterror(&message);
      die("Error processing command line options (%s)\n", message);
    }

    if (strcmp(option_name, "allow-weak-motifs") == 0) {
      options->allow_weak_motifs =  TRUE;
    } else if (strcmp(option_name, "bgfile") == 0) {
       options->bg_filename = option_value;
    } else if (strcmp(option_name, "e-thresh") == 0) {
      options->e_threshold = atof(option_value);
    } else if (strcmp(option_name, "fim") == 0) {
      options->fim = TRUE;
    } else if (strcmp(option_name, "gap-extend") == 0) {
      options->gap_extend = atof(option_value);
    } else if (strcmp(option_name, "gap-open") == 0) {
      options->gap_open = atof(option_value);
    } else if (strcmp(option_name, "global") == 0) {
      options->local_scoring = FALSE;
    } else if (strcmp(option_name, "maxseqs") == 0) {
      options->max_seq = atoi(option_value);
    } else if (strcmp(option_name, "model") == 0) {
      options->model_filename = option_value;
    } else if (strcmp(option_name, "motif") == 0) {
      add_string(option_value, options->requested_motifs);
    } else if (strcmp(option_name, "moitf-e-thresh") == 0) {
      options->motif_e_threshold = atof(option_value);
    } else if (strcmp(option_name, "moitf-p-thresh") == 0) {
      options->motif_p_threshold = atof(option_value);
    } else if (strcmp(option_name, "motif-pseudo") == 0) {
      options->motif_pseudo = atof(option_value);
    } else if (strcmp(option_name, "motif-scoring") == 0) {
      options->motif_scoring = TRUE;
    } else if (strcmp(option_name, "no-search") == 0) {
      options->perform_search = FALSE;
    } else if (strcmp(option_name, "nspacer") == 0) {
      options->spacer_states = atoi(option_value);
    } else if (strcmp(option_name, "order") == 0) {
      options->order_string = option_value;
    } else if (strcmp(option_name, "pam") == 0) {
      options->pam_distance = atoi(option_value);
    } else if (strcmp(option_name, "paths") == 0) {
      if (strcmp(option_value, "single") == 0) {
        options->paths = SINGLE;
      }
      else if (strcmp(option_value, "all") == 0) {
        options->paths = ALL;
      }
      else {
        fprintf(stderr, "Invalid path option: %s.\n", option_value);
        fprintf(stderr, usage);
        exit(1);
      }
    } else if (strcmp(option_name, "p-thresh") == 0) {
      options->p_threshold = atof(option_value);
      options->motif_scoring = TRUE;
    } else if (strcmp(option_name, "progress") == 0) {
      options->progress = atof(option_value);
    } else if (strcmp(option_name, "score-file") == 0) {
      options->score_filename = option_value;
    } else if (strcmp(option_name, "spacer-pseudo") == 0) {
      options->spacer_pseudo = atof(option_value);
    } else if (strcmp(option_name, "synth") == 0) {
      options->synth = TRUE;
    } else if (strcmp(option_name, "trans-pseudo") == 0) {
      options->trans_pseudo = atof(option_value);
    } else if (strcmp(option_name, "verbosity") == 0) {
      verbosity = (VERBOSE_T)atoi(option_value);
    } else if (strcmp(option_name, "zselo") == 0) {
      options->zselo = TRUE;
    }
  }

  // Read the required argument.
  if (option_index + 1 >= argc) {
    fprintf(stderr, "Missing required argument.\n");
    fprintf(stderr, usage);
    exit(1);
  }
  options->meme_filename = argv[option_index];
  option_index++;
  // If searching is turned on we have a second required argument
  if (options->perform_search == TRUE) {
    if (option_index + 1 != argc) {
      fprintf(stderr, "Missing required argument.\n");
      fprintf(stderr, usage);
      exit(1);
    }
    options->seq_filename = argv[option_index];
    option_index++;
  }
}

/*************************************************************************
 *  Entry point for program.
 *************************************************************************/
int main(int argc, char *argv[]) {

  OPTIONS_T options; 
  MHMM_T *the_hmm = NULL;

  process_command_line(argc, argv, &options);

  if (options.model_filename != NULL && file_exists(options.model_filename)) {
    // If a model file was specified and exists, we'll read the HMM from file.
    // Begin by assuming that it is an MHMM XML file.
    BOOLEAN_T result = read_mhmm_xml_file(
      options.model_filename, 
      &the_hmm
    );
    if (result == FALSE) {
      // It wasn't an MHMM XML file, try the old MHMM text format.
      read_mhmm_file(
        options.model_filename, 
        &the_hmm
      );
    }
    // Check that the HMM has linear topology
    if (the_hmm->type != LINEAR_HMM) {
      die(
        "beadstring requires a linear topology HMM.\n"
        "The HMM in %s has a %s topology.\n",
        options.model_filename,
        HMM_STRS[the_hmm->type]
      );
    }
  }
  else {
    // If a model file was not specifed or does not exist, we'll 
    // build the HMM from motif occurrences.
    the_hmm = build_linear_model(&options);
    if (options.model_filename != NULL) {
      // If a model file was specified and does not exit,
      // we'll write the HMM to the model file.
      write_mhmm_xml_to_file(the_hmm, options.model_filename);
    }
  }

  if (options.perform_search) {
    BOOLEAN_T got_evd = FALSE;
    search_with_model(the_hmm, &got_evd, &options);
    write_search_results(stdout, the_hmm, got_evd, &options);
  }

  /* Print the program parameters. */
  printf("Program parameters for beadstring\n");
  printf("  MEME file: %s\n", options.meme_filename);
  printf("  Motifs:");
  write_string_list(" ", options.requested_motifs, stdout);
  printf("\n");
  printf("  Model topology: linear\n");
  printf("  States per spacer: %d\n", options.spacer_states);
  printf(
    "  Spacers are free-insertion modules: %s\n", 
    options.fim ? "true" : "false"
  );
  printf("\n");

  free_mhmm(the_hmm);
  free_string_list(options.requested_motifs);

  return(0);
}

