/*************************************************************************
 * FILE: qvalue.c
 * AUTHOR: William Stafford Noble
 * CREATE DATE: 13 November 2007
 * PROJECT: MEME
 * COPYRIGHT: 2007, UW
 * DESCRIPTION: q value calculation, a la Benjamini Hochberg
 *************************************************************************/
#include "qvalue.h"
#include "array.h"
#include <stdio.h>

/*************************************************************************
 * Convert a set of p-values to a set of q-values.  The p-values must
 * be in sorted (ascending) order.  This function replaces them with
 * the corresponding q-values.
 *
 * The total_values parameter may be larger than the number of values
 * stored in the array of p-values.
 *************************************************************************/
void compute_qvalues
  (BOOLEAN_T compute_fdr,   // If true, only compute FDR, not q-value.
   ATYPE     pi_zero,       // Percentage of incorrect targets.
   int       total_values,  // Total number of values.
   ARRAY_T*  pvalues)
{
  // Make sure we got at least one p-value.
  int num_values = get_array_length(pvalues);
  if (num_values == 0) {
    return;
  }

  // Verify that the p-values are in the right range.
  int i_pvalue;
  for (i_pvalue = 0; i_pvalue < num_values; i_pvalue++) {
    ATYPE this_pvalue = get_array_item(i_pvalue, pvalues);
    myassert(1, this_pvalue >= 0.0, "Negative p-value (%g)\n", this_pvalue);
    myassert(1, this_pvalue <= 1.0, "p-value > 1 (%g)\n", this_pvalue);
  }

  // Verify that the p-values are in sorted order.
  ATYPE prev_pvalue = get_array_item(0, pvalues);
  for (i_pvalue = 1; i_pvalue < num_values; i_pvalue++) {
    ATYPE this_pvalue = get_array_item(i_pvalue, pvalues);
    myassert(1, this_pvalue >= prev_pvalue, "Unsorted p-values (%g < %g)\n",
	     this_pvalue, prev_pvalue);
    prev_pvalue = this_pvalue;
  }

  // Compute FDRs.
  for (i_pvalue = 0; i_pvalue < num_values; i_pvalue++) {
    ATYPE this_pvalue = get_array_item(i_pvalue, pvalues);
    ATYPE this_fdr = (pi_zero * this_pvalue * (float)total_values) 
      / (float)(i_pvalue + 1);
    set_array_item(i_pvalue, MIN(1.0,this_fdr), pvalues);
  }

  // No conversion if we only got one value or if only computing FDR.
  if ((num_values == 1) || (compute_fdr)) {
    return;
  }

  // Convert to q-values.
  prev_pvalue = get_array_item(num_values - 1, pvalues);
  for (i_pvalue = num_values - 2; i_pvalue >= 0; i_pvalue--) {
    ATYPE this_pvalue = get_array_item(i_pvalue, pvalues);
    if (prev_pvalue < this_pvalue) {
      set_array_item(i_pvalue, prev_pvalue, pvalues);
    }
    prev_pvalue = get_array_item(i_pvalue, pvalues);
  }
}

/*****************************************************************************
 * MAIN
 *****************************************************************************/
#ifdef MAIN

#include "simple-getopt.h"
VERBOSE_T verbosity = NORMAL_VERBOSE;

#define MAX_LINE 1000
#define MAX_PVALUES 100000

int main
  (int    argc,
   char * argv[])
{
  char * pvalue_filename = NULL;
  FILE * pvalue_file = NULL;

  int option_count = 1;
  int option_index = 0;
  char* option_name = NULL;
  char* option_value = NULL;
  const char *  message = NULL;
  char line[MAX_LINE];    // Buffer for reading.
  ARRAY_T* initial_pvalues = NULL;
  ARRAY_T* pvalues = NULL;
  ARRAY_T* qvalues = NULL;
  BOOLEAN_T compute_fdr = FALSE;  // By default, compute p-value.
  ATYPE pi_zero = 1;              // Percentage of incorrect targets.
  VERBOSE_T verbosity = NORMAL_VERBOSE;
  cmdoption const options[] = {
    { "pi-zero", REQUIRED_VALUE },
    { "fdr", NO_VALUE},
    { "verbosity", REQUIRED_VALUE }
  };

  // Define the usage message.
  char      usage[400] = "";
  strcat(usage, "USAGE: qvalue [options] <pvalues>\n");
  strcat(usage, "\n");
  strcat(usage, "   Options:\n");
  strcat(usage, "     --pi-zero <float>   (default = 1.0)\n");
  strcat(usage, "     --fdr\n");
  strcat(usage, "     --verbosity 1|2|3|4 (default = 2)\n");
  strcat(usage, "\n");

  /* Parse the command line. */
  simple_setopt(argc, argv, option_count, options);
  while(1) {
    // Read the next option, and break if we're done.
    int c = simple_getopt(&option_name, &option_value, &option_index);
    if (c == 0) {
      break;
    } else if (c < 0) {
      simple_getopterror(&message);
      die("Error processing command line options (%s)\n", message);
    }

    if (strcmp(option_name, "pi-zero") == 0) {
      pi_zero = atof(option_value);
      myassert(TRUE,
	       (pi_zero > 0.0) && (pi_zero <= 1.0),
	       "Invalid pi_zero (%g).\n", 
	       pi_zero);
    } else if (strcmp(option_name, "fdr") == 0) {
      compute_fdr = TRUE;
    } else if (strcmp(option_name, "verbosity") == 0) {
      verbosity = FALSE;
    }
  }

  // Read the single required argument.
  if (option_index + 1 != argc) {
    fprintf(stderr, usage);
    exit(1);
  }
  pvalue_filename = argv[option_index];

  // Read the p-values into a very large array.
  if (open_file(pvalue_filename, "r", TRUE, "p-values", "p-values",
		&pvalue_file) == 0) {
    exit(1);
  }
  initial_pvalues = allocate_array(MAX_PVALUES);
  int i_pvalue = 0;
  while (fgets(line, MAX_LINE, pvalue_file) != NULL) {
    ATYPE pvalue;
    if (sscanf(line, "%lf", &pvalue) != 1) {
      die("Can't parse line %d in %s (%s).\n", i_pvalue, pvalue_filename, 
	  line);
    }
    if (i_pvalue >= MAX_PVALUES) {
      die("Can't read more than %d p-values.\n", MAX_PVALUES);
    }
    set_array_item(i_pvalue, pvalue, initial_pvalues);
    i_pvalue++;
  }
  fclose(pvalue_file);
  int num_pvalues = i_pvalue;
  if (verbosity >= NORMAL_VERBOSE) {
    fprintf(stderr, "Read %d p-values from %s.\n", num_pvalues, 
	    pvalue_filename);
  }

  // Make a smaller array to contain just the pvalues.
  pvalues = allocate_array(num_pvalues);
  for (i_pvalue = 0; i_pvalue < num_pvalues; i_pvalue++) {
    set_array_item(i_pvalue,
		   get_array_item(i_pvalue, initial_pvalues),
		   pvalues);
  }

  // Sort it.
  sort_array(FALSE, pvalues);

  // Make a copy of it.
  qvalues = allocate_array(num_pvalues);
  copy_array(pvalues, qvalues);

  // Compute q-values.
  compute_qvalues(compute_fdr, pi_zero, get_array_length(qvalues), qvalues);

  // Print to stdout.
  for (i_pvalue = 0; i_pvalue < num_pvalues; i_pvalue++) {
    printf("%g\t%g\n", 
	   get_array_item(i_pvalue, pvalues),
	   get_array_item(i_pvalue, qvalues));
  }

  return(0);
}

#endif
