/********************************************************************
 * FILE: motiph.c
 * AUTHOR: William Stafford Noble, Charles E. Grant
 * CREATE DATE: 12/17/2004
 * PROJECT: EVOMCAST
 * COPYRIGHT: 2004, UW
 ********************************************************************/

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "alignment.h"
#include "alphabet.h"
#include "clustalw-io.h"
#include "evomodel.h"
#include "fasta-io.h"
#include "meme-io.h"
#include "mhmm-state.h"
#include "motif.h"
#include "motiph-scoring.h"
#include "scored-sites.h"
#include "pssm.h"
#include "seq.h"
#include "simple-getopt.h"
#include "tree.h"

// Nucleotide alphabest order as in motif.h
extern char alphabet[];

char* program_name = NULL;
VERBOSE_T verbosity = NORMAL_VERBOSE;

static TREE_T* trim_tree_to_alignment(TREE_T* tree, ALIGNMENT_T* alignment);
static ARRAY_T* read_bg_from_file(char* bg_filename);
static ARRAY_T* read_bg_from_alignment(
  ALIGNMENT_T* alignment, 
  MODEL_TYPE_T model_type
);

/*************************************************************************
 * Entry point for shadow
 *************************************************************************/
int main(int argc, char *argv[]) {

  BOOLEAN_T use_selected_motif_index = FALSE;
  BOOLEAN_T use_file_list = FALSE;
  char* bg_filename = NULL;
  char* motif_name = "motif"; // Use this motif name in the output.
  char* seq_name = NULL;      // Use this sequence name in the output.
  int selected_motif_index = 0;
  double pthresh = 1e-6;
  //const double default_pv_pthresh = 1e-6;
  double gap_cost = 0.0;
  double fg_rate = 1.0;
  double bg_rate = 1.0;
  double purine_pyrimidine = 1.0; // r
  double transition_transversion = 0.5; // R
  GAP_SUPPORT_T gap_support = SKIP_GAPS;
  MODEL_TYPE_T model_type = F81_MODEL;
  char* gff_filename = NULL;
  program_name = "shadow";

  /**********************************************
   * COMMAND LINE PROCESSING
   **********************************************/

  // Define command line options. (FIXME: Repeated code)
  const int num_options = 14;
  cmdoption const shadow_options[] = {
    {"list", NO_VALUE},
    {"model", REQUIRED_VALUE},
    {"pthresh", REQUIRED_VALUE},
    {"pur-pyr", REQUIRED_VALUE},
    {"transition-transversion", REQUIRED_VALUE},
    {"bg", REQUIRED_VALUE},
    {"fg", REQUIRED_VALUE},
    {"gap", REQUIRED_VALUE},
    {"gap-cost", REQUIRED_VALUE},
    {"gap-freq", REQUIRED_VALUE},
    {"bgfile", REQUIRED_VALUE},
    {"sequence-name", REQUIRED_VALUE},
    {"gff", REQUIRED_VALUE},
    {"verbosity", REQUIRED_VALUE}
  };

  int option_index = 0;

  // Define the usage message.
  char usage[1000] = "";
  strcat(usage, "USAGE: shadow [options] <alignment> <tree>\n");
  strcat(usage, "\n");
  strcat(usage, "   Options:\n");

  // Evolutionary model parameters.
  strcat(usage, "     --list\n");
  strcat(usage, "     --model [single|average|jc|k2|f81|f84|hky|tn]");
  strcat(usage, " (default=f81)\n");
  strcat(usage, "     --pur-pyr <float> (default=1.0)\n");
  strcat(usage, "     --transition-transversion <float> (default=0.5)\n");
  strcat(usage, "     --bg <float> (default=1.0)\n");
  strcat(usage, "     --fg <float> (default=1.0)\n");
  strcat(usage, "     --gap skip | fixed | wildcard | minimum");
  strcat(usage, " (default=skip)\n");
  strcat(usage, "     --gap-cost <float> (default=0.0)\n");

  // Miscellaneous parameters
  strcat(usage, "     --pthresh <float> (default 1e-6)\n");
  strcat(usage, "     --bgfile <background> (default from alignment)\n");
  strcat(usage, "     --sequence-name <string> (default first sequence" \
                " in alignment)\n");
  strcat(usage, "     --gff <filename>\n");
  strcat(usage, "     --verbosity [1|2|3|4] (default 2)]\n");
  strcat(usage, "\n");


  // Parse the command line.
  if (simple_setopt(argc, argv, num_options, shadow_options) != NO_ERROR) {
    die("Error processing command line options: option name too long.\n");
  }

  while (TRUE) { 
    int c = 0;
    char* option_name = NULL;
    char* option_value = NULL;
    const char * message = NULL;

    // Read the next option, and break if we're done.
    c = simple_getopt(&option_name, &option_value, &option_index);
    if (c == 0) {
      break;
    } else if (c < 0) {
      (void) simple_getopterror(&message);
      die("Error processing command line options (%s)\n", message);
    }
    
    if (strcmp(option_name, "model") == 0) {
      if (strcmp(option_value, "jc") == 0) {
        model_type = JC_MODEL;
      } else if (strcmp(option_value, "k2") == 0) {
        model_type = K2_MODEL;
      } else if (strcmp(option_value, "f81") == 0) {
        model_type = F81_MODEL;
      } else if (strcmp(option_value, "f84") == 0) {
        model_type = F84_MODEL;
      } else if (strcmp(option_value, "hky") == 0) {
        model_type = HKY_MODEL;
      } else if (strcmp(option_value, "tn") == 0) {
        model_type = TAMURA_NEI_MODEL;
      } else if (strcmp(option_value, "single") == 0) {
        model_type = SINGLE_MODEL;
      } else if (strcmp(option_value, "average") == 0) {
        model_type = AVERAGE_MODEL;
      } else {
        die("Unknown model: %s\n", option_value);
      }
    } else if (strcmp(option_name, "pur-pyr") == 0){
        purine_pyrimidine = atof(option_value);
    } else if (strcmp(option_name, "transition-transervsion") == 0){
        transition_transversion = atof(option_value);
    } else if (strcmp(option_name, "bg") == 0){
      bg_rate = atof(option_value);
    } else if (strcmp(option_name, "fg") == 0){
      fg_rate = atof(option_value);
    } else if (strcmp(option_name, "pthresh") == 0){
      pthresh = atof(option_value);
    } else if (strcmp(option_name, "gap") == 0){
      if (strcmp(option_value, "skip") == 0) {
        gap_support = SKIP_GAPS;
      } else if (strcmp(option_value, "fixed") == 0) {
        gap_support = FIXED_GAP_COST;
      } else if (strcmp(option_value, "wildcard") == 0) {
        gap_support = WILDCARD_GAP;
      } else if (strcmp(option_value, "minimum") == 0) {
        gap_support = MIN_GAPS;
      } else {
        die("Unknown gap handling method: %s\n", option_value);
      }
    } else if (strcmp(option_name, "gap-cost") == 0) {
      gap_cost = atof(option_value);
    } else if (strcmp(option_name, "motif") == 0){
        use_selected_motif_index = TRUE;
        selected_motif_index = atoi(option_value) - 1;
    } else if (strcmp(option_name, "motif-name") == 0){
        motif_name = option_value;
    } else if (strcmp(option_name, "list") == 0){
      use_file_list = TRUE;
    } else if (strcmp(option_name, "bgfile") == 0){
      bg_filename = option_value;
    } else if (strcmp(option_name, "sequence-name") == 0){
        seq_name = option_value;
    } else if (strcmp(option_name, "gff") == 0) {
      gff_filename = option_value;
    } else if (strcmp(option_name, "verbosity") == 0){
        verbosity = atoi(option_value);
    }
  }

  // Must have tree and alignment file names
  if (argc != option_index + 2) {
    fprintf(stderr, usage);
    exit(EXIT_FAILURE);
  } 

  set_alphabet(verbosity, "ACGT");

  /****************************************************
   * Read the names of the alignment files.
   ****************************************************/
  STRING_LIST_T* filenames = NULL;
  int num_filenames = 0;
  if (use_file_list) {
    filenames = read_string_list_from_file(argv[option_index]);
    num_filenames = get_num_strings(filenames);
  }
  else {
    filenames = new_string_list();
    add_string(argv[option_index], filenames);
    num_filenames = 1;
  }
  option_index++;

  /**********************************************
   * Read the phylogenetic tree.
   **********************************************/
  char* tree_filename = NULL;
  TREE_T* tree = NULL;
  if (MOTIPH || SHADOW) {
    tree_filename = argv[option_index];
    option_index++;
    tree = read_tree_from_file(tree_filename);
  }

  /**********************************************
   * Determine the background frequencies if 
   * background file provided.
   **********************************************/
  ARRAY_T* bg_freqs = NULL;
  if (bg_filename != NULL) {
    bg_freqs = read_bg_from_file(bg_filename);
    if (verbosity >= NORMAL_VERBOSE) {
      fprintf(stderr, "Read background frequencies from %s.\n", bg_filename);
    }
  }

  /****************************************************
   * Open the optional GFF file for output.
   ****************************************************/
  FILE* gff_file = NULL;
  if (gff_filename != NULL) {
    if (open_file(gff_filename, "w", FALSE, "GFF", "GFF", &gff_file) == 0) {
      exit(1);
    }
  }

  /****************************************************
   * Do the scoring.
   ****************************************************/
  int window_size = 1;
  const int num_models = 2;
  EVOMODEL_T* models[num_models];
  ALIGNMENT_T* alignment = NULL;

  int num_alignment_species = 0;
  int num_tree_species = 0;
  STRING_LIST_T* alignment_species = NULL;
  STRING_LIST_T* tree_species = NULL;

  SCORED_SITES_T* site_list;
  site_list = new_scored_sites();

  // Score the reference sequence in the alignment using these models.
  int file_index;
  for(file_index = 0; file_index < num_filenames; file_index++) {

    char* filename = get_nth_string(file_index, filenames);
    alignment = read_alignment_from_clustalw_file(filename);
    //TREE_T* trimmed_tree = trim_tree_to_alignment(tree, alignment);
    // For now we require that the species in the alignment 
    // exactly match the species in the full tree.
    // FIXME Eventually we should trim the tree of those species
    // not in the aligment, and cache the PSSM for the trimmed
    // tree and alignment.
    num_alignment_species = get_num_aligned_sequences(alignment);
    alignment_species = get_species_names(alignment);
    sort_string_list(alignment_species);
    num_tree_species = leaf_count(tree);
    tree_species = make_leaf_list(tree);
    sort_string_list(tree_species);
    if (num_tree_species != num_alignment_species) {
      fprintf(stderr, "Skiping alignment %s, %d species in alignment"
        " but %d species in tree.", filename, num_alignment_species,
        num_tree_species);
      break;
    }
    int species_index = 0;
    for(species_index = 0; species_index < num_alignment_species; 
      species_index++) {
      char* alignment_spec = get_nth_string(species_index, alignment_species);
      char* tree_spec = get_nth_string(species_index, tree_species);
      if (strcmp(alignment_spec, tree_spec) != 0) {
        fprintf(stderr, "Skiping alignment %s, species %s in alignment"
          " but not in tree.", filename, alignment_spec);
        break;
      }
    }

    if (bg_filename == NULL) {
      // Take background freq. from alignment if no background file specified.
      bg_freqs = read_bg_from_alignment(alignment, model_type);
      if (verbosity >= NORMAL_VERBOSE) {
        fprintf(stderr, "Read background frequencies from alignment.\n");
      } 
    }

    // Build the background model using the equilibrium frequencies
    // from the alignment and the background mutation rate.
    models[0] = make_model(
      model_type,
      bg_rate,
      transition_transversion,
      purine_pyrimidine,
      bg_freqs,
      FALSE
    );

    // Build the foreground model using the equilibrium frequencies
    // from the alignment and the foreground mutation rate.
    models[1] = make_model(
      model_type,
      fg_rate,
      transition_transversion,
      purine_pyrimidine,
      bg_freqs,
      FALSE
    );

    // Build PSSM for motiph
    MATRIX_T* pssm = build_alignment_pssm(
      alignment_species,
      window_size + 1, 
      models, 
      tree, 
      gap_support
    );

    // The first row of the matrix contains the  probabilities for 
    // the alignment columns given the background model
    ARRAY_T* alignment_col_probs = allocate_array(get_num_cols(pssm)); 
    copy_array(get_matrix_row(0, pssm), alignment_col_probs);
    remove_matrix_row(0, pssm);

    // Build tables to translate log-odds scores to p-values
    ARRAY_T* pv_lookup = NULL;
    scale_pssm(pssm, 100);
    pv_lookup = get_pv_lookup(
      pssm, 
      100,
      get_num_cols(pssm), 
      alignment_col_probs 
    );

    // Build a table for converting the index into the alignment 
    // to an index into ungapped reference sequence.
    int* coord_conv_table = make_alignment_to_seq_table( 0, alignment);

    score_sequence_in_alignment(
      0, 
      alignment, 
      "conservation", 
      NULL, 
      seq_name, 
      window_size,
      models, 
      pssm,
      pv_lookup,
      coord_conv_table,
      gap_support, 
      gap_cost,
      pthresh,
      site_list
    );

    free_string_list(alignment_species);
    free_string_list(tree_species);
    free_array(alignment_col_probs);
    free_array(pv_lookup);
    free_matrix(pssm); 
    free_alignment(alignment);
    if (models != NULL) {
      int model_index;
      for (model_index = 0; model_index < (window_size + 1); model_index++) {
        free_model(models[model_index]);
      }
    }
  }

  // Print the sites.
  print_scored_sites(program_name, site_list, gff_file, stdout);
  free_scored_sites(site_list);

  free_array(bg_freqs);
  free_tree(TRUE, tree);
  free_string_list(filenames);

  return(0);
}

/*******************************************************************
  Read the background nucleotide frequenices from a file
 ********************************************************************/
ARRAY_T* read_bg_from_file(char* bg_filename) {

  FILE* bg_file = NULL;
  if (open_file(bg_filename, "r", TRUE, "background", 
                "background nucleotide frequencies", &bg_file) == 0) {
    die("Couldn't open the file %s.\n", bg_filename);
  }
  
  // Read four nucleotide frequencies from file
  const int bufsize = 80;
  char line[bufsize];
  char word[bufsize];
  int num_read = 0;
  float a_freq = 0.0;
  float c_freq = 0.0;
  float t_freq = 0.0;
  float g_freq = 0.0;
  while (TRUE) {
    if (fgets(line, 80, bg_file) == NULL) {
      break;
    }
    if (sscanf(line, "# %s", word) == 1) {
      continue;
    }
    if (sscanf(line, "A %f", &a_freq) == 1) {
      num_read++;
      continue;
    }
    if (sscanf(line, "C %f", &c_freq)) {
      num_read++;
      continue;
    }
    if (sscanf(line, "G %f", &g_freq) == 1) {
      num_read++;
      continue;
    }
    if (sscanf(line, "T %f", &t_freq) == 1) {
      num_read++;
      continue;
    }
    // Once we've read the four frequences we don't
    // need to read any more.
    if (num_read == 4) {
      break;
    }
  }

  fclose(bg_file);

  if (num_read != 4) {
    die("Can't read background frequences from file %s.\n", bg_filename);
  }

  int alph_size = get_alph_size(ALPH_SIZE);
  ARRAY_T* bg_freqs = allocate_array(alph_size);
  int index = 0;
  index = alphabet_index('A', alphabet);
  set_array_item(index, a_freq, bg_freqs);
  index = alphabet_index('C', alphabet);
  set_array_item(index, c_freq, bg_freqs);
  index = alphabet_index('G', alphabet);
  set_array_item(index, g_freq, bg_freqs);
  index = alphabet_index('T', alphabet);
  set_array_item(index, t_freq, bg_freqs);

  return bg_freqs;
}

/*******************************************************************
  Determine the background nucleotide frequenices from an alignment.
 ********************************************************************/
static ARRAY_T* read_bg_from_alignment(
  ALIGNMENT_T* alignment, 
  MODEL_TYPE_T model_type
)
{
  // Set up the frequency array.
  ARRAY_T* bg_freqs;
  if (model_type == SINGLE_MODEL) {
    SEQ_T* seq = get_alignment_sequence(0, alignment);
    bg_freqs = get_sequence_freqs(seq);
    free_seq(seq);
    if (verbosity >= NORMAL_VERBOSE) {
      fprintf(stderr, "Computing background frequencies from %s.\n",
              get_seq_name(seq));
    }
  } else {
    bg_freqs = get_alignment_freqs(alignment);
    if (verbosity >= NORMAL_VERBOSE) {
      fprintf(stderr, "Computing background frequencies from alignment.\n");
    }
  }
  if (verbosity >= NORMAL_VERBOSE) {
    fprintf(stderr, "Background frequencies: ");
    print_array(bg_freqs, 6, 3, TRUE, stderr);
  }

  return bg_freqs;
}

TREE_T* trim_tree_to_alignment(TREE_T* tree, ALIGNMENT_T* alignment) {

  // Trim species from the tree that aren't in the alignment.
  STRING_LIST_T* species_names = get_species_names(alignment);
  TREE_T* trimmed_tree = trim_tree(TRUE, tree, species_names);
  free_string_list(species_names);
  if (verbosity >= HIGH_VERBOSE) {
    fprintf(stderr, "Trimmed tree: ");
    write_tree(trimmed_tree, stderr);
  }

  return trimmed_tree;

}
